-------------------------------------------------------------------------------------
-- | Namespace Definitions |---------------------------------------------------------
-------------------------------------------------------------------------------------

-- Retrieve the addon name and the namespace table.
local addon, npcg = ...

local functua 	= npcg.functua
local util 			= npcg.functua.util

local generator = { }
local data 			= { }
local tests 		= { }

-- generators and data do not exist yet
npcg.generators.class	= generator
npcg.data.class				= data
npcg.tests.class			= tests

---------------------------------------------------------------------------------------------------------------------------------------
-- | Raw Data | -----------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- defaultTraits :: { TraitDepencency } -> { TraitDepencency }
-- Provides defaultTraits for trait dependencies, so others don't have to fill them in.
local function defaultTraits(replacements)
  local replacements = replacements or { }
  return {
    classTypes  = replacements["classTypes"] or {       },
    races       = replacements[     "races"] or npcg.data.race.allRacesExcept("illidari", "death knight", "void elf"),
		tribes			= replacements[    "tribes"] or npcg.data.tribe.allTribesExcept("lightforged"),
    faiths      = replacements[    "faiths"] or { "any" },
    genders     = replacements[   "genders"] or { "any" },
    alignments  = replacements["alignments"] or { "any" },
    ages        = replacements[      "ages"] or npcg.data.age.olderThan("teen"),
		tags				= replacements[      "tags"] or {       },
    weight      = 1
  }
end

-- TODO: Rename
-- Data Table, this will be loaded in externally. Do not use this directly, use the functions instead.
local defaultTable = {
-- a
[ "adventurer"            ] = defaultTraits({
  classTypes = {
    "explorer",
    "mercenary", },
	alignments = npcg.data.alignment.alignmentsByEthics("chaotic")
}),
[ "alchemist"             ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[	"anchorite"							] = defaultTraits({
	classTypes = {
		"priest",
	},
	races = {
		"draenei",
	},
	tribes = {
		"any",
	},
}),
[ "animalhandler"         ] = defaultTraits({
  classTypes = {
    "animal",
    "hunter", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "antiquities dealer"    ] = defaultTraits({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "adult" )
}),
[ "archer"                ] = defaultTraits({
  classTypes = {
    "military",
    "hunter",
    "mercenary",
    },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" )
}),
[ "archeologist"          ] = defaultTraits({
  classTypes = {
    "scholar",
    "explorer", },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "\"archmage\""          ] = defaultTraits({
  classTypes = {
    "playing", "mage" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "high elf",
    "human",
    "night elf",
		"nightborne",
		},
  ages = npcg.data.age.agesBetween( "child", "teen" )
}),
[ "armorer"               ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "young adult" ),
}),
[	"artificer"							] = defaultTraits({
	classTypes = {
		"military",
		"inventor",	-- "Construct, Engineer, and Create Artifacts. Shape Gems and Crystals.
		"light",
	},
	races = {
		"draenei",
	},
	tribes = {
		"lightforged",
	},
}),
[ "assassin"              ] = defaultTraits({
  classTypes = {
    "rogue",
    "covert", },
  races = npcg.data.race.allRacesExcept(
		"draenei",
		"tauren",
		"death knight",
		"illidari" ),
	alignments = npcg.data.alignment.alignmentsByEthics("chaotic"),
}),
[ "astrologer"            ] = defaultTraits({
  classTypes = {
    "scholar" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "high elf",
    "human",
    "night elf",
		"nightborne",
		"void elf", },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "astronomer"            ] = defaultTraits({
  classTypes = {
    "scholar" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "high elf",
    "human",
    "night elf",
		"nightborne",
		"void elf"		},
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "author"                ] = defaultTraits({
  classTypes = {
    "scholar", },
}),
-- b
[ "baker"                 ] = defaultTraits({
  classTypes = {
    "food",
    "trader", },
}),
[ "bard"                  ] = defaultTraits({
  classTypes = {
    "entertainer",
    "bar", },
}),
[ "bartender"             ] = defaultTraits({
  classTypes = {
    "bar", },   
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" )
}),
[	"battlemage"						] = defaultTraits({
	classTypes = {
		"mage",
		"military",
	},
	races = {
		"blood elf",
		"gnome",
		"human",
		"high elf",
		"nightborne",
		"void elf",
	},
  ages = npcg.data.age.agesBetween( "young adult", "adult" )
}),
[ "beggar"                ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "blacksmith"            ] = defaultTraits({
  classTypes = {
    "crafter" },
  ages = npcg.data.age.agesBetween( "young adult", "adult" )
}),
[ "brat"                  ] = defaultTraits({
  classTypes = {
    "playing" },  
  ages = npcg.data.age.agesBetween( "child", "child" ),
}),
[ "brawler"               ] = defaultTraits({
  classTypes = {
    "bar",
    "mercenary", },
  races = npcg.data.race.allRacesExcept(
    "pandaren", "illidari", "death knight", "void elf"), 
  ages = npcg.data.age.olderThan( "teen" ),
	alignment = npcg.data.alignment.alignmentsByEthics("chaotic"),
}),
[ "bloodknight"           ] = defaultTraits({
  classTypes = {
    "paladin",
		"town support", },
  races = {
    "blood elf" },
  faiths = npcg.data.faith.faithsByTag( "light" ),
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful good",
		"neutral good",
	}
}),
[ "blood mage"            ] = defaultTraits({
  classTypes = {
    "mage",
    "scholar", },
  races = {
    "blood elf" },  
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "bodyguard"             ] = defaultTraits({
  classTypes = {
    "mercenary", },   
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "bookbinder"            ] = defaultTraits({
  classTypes = {
    "crafter",
    "scholar" },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "forsaken",
    "worgen" },  
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "bounty hunter"         ] = defaultTraits({
  classTypes = {
    "mercenary",
    "rogue", },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = npcg.data.alignment.alignmentsByEthics("chaotic"),
}),
[ "bowyer"                ] = defaultTraits({
  classTypes = {
    "crafter",
    "hunter" },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" )
}),
[ "brewer"                ] = defaultTraits({
  classTypes = {
    "crafter", },
}),
[ "brothelkeeper"         ] = defaultTraits({
  classTypes = {
    "illegal", },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "orc",
    "pandaren",
    "troll",
    "forsaken",
    "worgen" },
  ages = npcg.data.age.olderThan("young adult"),
	alignments = {
		"chaotic neutral",
		"chaotic evil",
		"neutral evil",
	},
}),
[ "brothelworker"         ] = defaultTraits({
  classTypes = {
    "illegal",
    "rogue" },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "nightborne",
    "orc",
    "pandaren",
    "troll",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "burglar"               ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.agesBetween( "teen", "adult" ),
	alignments = {
		"chaotic evil",
		"neutral evil",
	}
}),
[ "butcher"               ] = defaultTraits({
  classTypes = {
    "food",
    "trader", },
  ages = npcg.data.age.olderThan( "teen" )
}),
-- c
[ "candlemaker"           ] = defaultTraits({
  classTypes = {
    "crafter", },
  races = {
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "pandaren",
    "forsaken",
    "worgen" },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "carpenter"             ] = defaultTraits({
  classTypes = {
    "crafter", },
}),
[ "carpenter apprentice"  ] = defaultTraits({
  classTypes = {
    "crafter",
    "apprentice" },
  ages = {
		"teen" },
}),
[ "champion"							] = defaultTraits({
	classTypes = {
		"military" },
	races = { "any" },
	age = npcg.data.age.olderThan("adult"),
}),
[ "clergy"                ] = defaultTraits({
  classTypes = {
    "priest",
    "light", },
  faiths = npcg.data.faith.faithsByTag( "light" ),
  ages = npcg.data.age.olderThan( "teen" ),
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
	}
}),
[ "clothier"              ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "cobbler"               ] = defaultTraits({
  classTypes = {
    "crafter", },
  races = npcg.data.race.allRacesExcept(
    "draenei",
    "pandaren",
    "tauren",
		"illidari",
		"death knight",
		"void elf"),  
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "cobbler apprentice"    ] = defaultTraits({
  classTypes = {
    "crafter",
    "apprentice" },
  ages = npcg.data.age.agesBetween( "child","teen" )
}),
[ "commoner"              ] = defaultTraits({
  classTypes = {
    "citizen", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "cook"                  ] = defaultTraits({
  classTypes = {
    "food",
    "bar", },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" )
}),
[ "craftsman"             ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "cultist"               ] = defaultTraits({
  classTypes = {
    "priest",
    "dark" },
  faiths = npcg.data.faith.faithsByTag( "cult" ), 
  ages = npcg.data.age.olderThan( "child" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("evil"),
}),
-- d
[ "dark ranger"           ] = defaultTraits({
  classTypes = {
    "hunter",
		"military", },
  races = {
    "forsaken" },
	tribes = {
		"blood elf",
		"high elf",
	},
	genders = {
		"female",
	},
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful neutral",
	},
}),
[ "demon hunter"					] = defaultTraits({
	classTypes = {
		"military",
	},
	races = {
		"illidari" },
}),
[ "diplomat"              ] = defaultTraits({
  classTypes = {
    "town support",
		"diplomacy", },
  ages = npcg.data.age.olderThan( "young adult" ),
	races = { "any" }
}),
[ "doctor"                ] = defaultTraits({
  classTypes = {
    "medical" },
  ages = npcg.data.age.olderThan( "adult" ),
	alignments = {
		"lawful good",
		"neutral good",
		"chaotic good",
		"lawful neutral",
	}
}),
[ "doomsayer"             ] = defaultTraits({
  classTypes = {
    "town support",
    "dark" },
  faiths = npcg.data.faith.faithsByTag( "cult" ),
  ages = npcg.data.age.olderThan( "child" ),
	alignments = {
		"lawful evil",
		"neutral evil",
	}
}),
[ "druid"                 ] = defaultTraits({
  classTypes = {
    "druid",
		"military", },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" ),
	alignments = {
		"lawful good",
		"neutral good",
		"true neutral",
		"lawful neutral",
	}
}),
[ "drummer"               ] = defaultTraits({
  classTypes = {
    "entertainer" },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "drunk"                 ] = defaultTraits({
  classTypes = {
    "illegal",
    "town support", },
}),
-- e
[ "enchanter"             ] = defaultTraits({
  classTypes = {
    "mage",
    "crafter" },
}),
[ "enforcer"							] = defaultTraits({
	classTypes = {
		"military",
	},
	races = {
		"illidari",	},
}),
[ "engineer"              ] = defaultTraits({
  classTypes = {
    "crafter",
    "inventor", },
}),
[ "entrepreneur"          ] = defaultTraits({
  classTypes = {
    "trader" },
}),
[ "executioner"						] = defaultTraits({
	races = {
		"blood elf",
		"death knight",
		"dwarf",
		"forsaken",
		"goblin",
		"half elf",
		"high elf",
		"human",
		"illidari",
		"nightborne",
		"troll",
		"worgen",
	},
	ages = npcg.data.age.olderThan("young adult"),
}),
[ "explorer"              ] = defaultTraits({
  classTypes = {
    "explorer" },
  ages = npcg.data.age.olderThan( "teen" ),
	alignments = npcg.data.alignment.alignmentsByEthics("chaotic"),
}),
-- f
[ "farstrider"            ] = defaultTraits({
  classTypes = {
    "military", "hunter" },
  races = {
    "blood elf" },
  ages = npcg.data.age.agesBetween( "adult", "mature adult" ),
	alignment = {
		"chaotic good",
		"neutral good",
		"chaotic neutral",
		"true neutral",
	}
}),
[ "farmer"                ] = defaultTraits({
  classTypes = {
    "gatherer" },
  ages = npcg.data.age.agesBetween( "young adult", "elder" )
}),
[ "fisherman"             ] = defaultTraits({
  classTypes = {
    "gatherer",
    "harbor", },
  ages = npcg.data.age.agesBetween( "young adult", "adult" )
}),
[ "fishmonger"            ] = defaultTraits({
  classTypes = {
    "trader" },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "fruit vendor"          ] = defaultTraits({
  classTypes = {
    "trader" },   
  ages = npcg.data.age.olderThan( "child" )
}),
-- g
[ "gang member"           ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
	alignments = {
		"chaotic neutral",
		"lawful neutral",
		"neutral evil",
		"lawful evil",
	}
}),
[ "gardener"              ] = defaultTraits({
  classTypes = {
    "town support" },
  races = {
    "blood elf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "worgen" },
  ages = npcg.data.age.olderThan( "child" )
}),
[	"ghostblade"						] = defaultTraits({
	races = {
		"void elf" },
	classTypes = {
		"rogue",
		"military",
	},
}),
[ "grunt"                 ] = defaultTraits({
  classTypes = {
    "town support",
    "mercenary", },
  races = {
    "orc" },  
  ages = npcg.data.age.agesBetween( "young adult", "adult" )
}),
[ "guard"                 ] = defaultTraits({
  classTypes = {
    "town support" },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
	},
}),
-- h
[ "hairdresser"           ] = defaultTraits({
  classTypes = {
    "town support" },   
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "hay merchant"          ] = defaultTraits({
  classTypes = {
    "trader" },   
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "headhunter"            ] = defaultTraits({
  classTypes = {
    "mercenary", "hunter" },
  races = {
    "troll" },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "herald"                ] = defaultTraits({
  classTypes = {
    "citizen" },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "orc",
    "pandaren",
    "worgen" },
}),
[ "herbalist"             ] = defaultTraits({
  classTypes = {
    "gatherer" },   
}),
[ "homeless person"       ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages =  {
		"any",
	},
}),
[ "horsetrainer"          ] = defaultTraits({
  classTypes = {
    "animal" },
  races = {
  "blood elf",
  "dwarf",
  "half elf",
  "half orc",
  "high elf",
  "human",
  "forsaken",
  "worgen" },
}),
[ "hunter"                ] = defaultTraits({
  classTypes = {
    "hunter",
		"military", },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" )
}),
-- i
[ "illuminator"           ] = defaultTraits({
  classTypes = {
    "town support", },
  races = {
    "blood elf",
    "dwarf",
    "goblin",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "orc",
    "pandaren",
    "tauren",
    "worgen",
    "forsaken" },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "innkeeper"             ] = defaultTraits({
  classTypes = {
    "bar",
    "citizen", },
  ages = npcg.data.age.olderThan( "adult" )
}),
[ "inventor"              ] = defaultTraits({
  classTypes = {
    "inventor",
    "trader", },
  ages = npcg.data.age.olderThan( "young adult" )
}),
-- j
[ "jailer"                ] = defaultTraits({
  classTypes = {
    "town support" },
  ages = npcg.data.age.olderThan( "young adult" ),
}),
[ "jester"                ] = defaultTraits({
  classTypes = {
    "entertainer" },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "jewelcrafter"          ] = defaultTraits({
  classTypes = {
    "crafter",
    "trader" },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "juggler"               ] = defaultTraits({
  classTypes = {
    "entertainer" },   
  ages = npcg.data.age.agesBetween("child", "elder")
}),
-- k
[ "kitchen help"          ] = defaultTraits({
  classTypes = {
    "bar",
    "food", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "knight"                ] = defaultTraits({
  classTypes = {
    "paladin",
    "military" },
  races = {
    "human" },  
  ages = npcg.data.age.olderThan( "adult" ),
	alignments = {
		"lawful good",
		"lawful neutral"
	}
}),
-- l
[ "laborer"               ] = defaultTraits({
  classTypes = {
    "citizen" },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" )
}),
[ "leatherworker"         ] = defaultTraits({
  classTypes = {
    "crafter", "hunter" },
}),
[ "loan provider"         ] = defaultTraits({
  classTypes = {
    "town support" },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful neutral",
	}
}),
[	"locus researcher"			] = defaultTraits({
	races = {
		"void elf" },
	classTypes = {
		"mage",
		"scholar", },
}),
[ "lookout"               ] = defaultTraits({
  classTypes = {
    "rogue",
    "mercenary" },
  ages = npcg.data.age.agesBetween( "child", "young adult" ),
	races = { "any" },
}),
[ "lorewalker"            ] = defaultTraits({
  classTypes = {
    "scholar",
    "explorer", },
  races = {
    "pandaren" },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful good",
		"lawful neutral",
	}
}),
[ "lumberjack"            ] = defaultTraits({
  classTypes = {
    "gatherer" },   
  ages = npcg.data.age.agesBetween( "teen", "mature adult" )
}),
-- m
[ "mage"                  ] = defaultTraits({
  classTypes = {
    "mage",
    "scholar", },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" )
}),
[ "mathmatician"          ] = defaultTraits({
  classTypes = {
    "scholar" },
  races = {
    "blood elf",
    "dwarf",
		"forsaken",
    "gnome",
    "goblin",
    "half elf",
    "high elf",
    "human",

    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful neutral",
	}
}),
[ "medic"                 ] = defaultTraits({
  classTypes = {
    "medical" },
  ages = npcg.data.age.agesBetween( "young adult", "elder" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("good"),
}),
[ "mercenary"             ] = defaultTraits({
  classTypes = {
    "mercenary",
    "warrior", },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("neutral"),
}),
[ "merchant"              ] = defaultTraits({
  classTypes = {
    "trader" },
  ages = npcg.data.age.agesBetween( "teen", "elder" )
}),
[ "messenger"             ] = defaultTraits({
  classTypes = {
    "citizen",
		"diplomacy", },
  ages = npcg.data.age.olderThan( "child" ),
	races = { "any" },
}),
[ "military advisor"      ] = defaultTraits({
  classTypes = {
    "military", },
  ages = npcg.data.age.olderThan( "adult" )
}),
[ "mime"                  ] = defaultTraits({
  classTypes = {
    "entertainer" },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "worgen" },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "miner"                 ] = defaultTraits({
  classTypes = {
    "gatherer" },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" )
}),
[ "monk"                  ] = defaultTraits({
  classTypes = {
    "monk" },
  ages = npcg.data.age.olderThan( "teen" ),
	alignments = {
		"lawful good",
		"lawful neutral",
		"true neutral",
	}
}),
[ "musician"              ] = defaultTraits({
  classTypes = {
    "entertainer" },
  ages = npcg.data.age.olderThan( "teen" )
}),
-- n
[ "nanny"                 ] = defaultTraits({
  classTypes = {
    "children",
    "citizen", },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "navigator"             ] = defaultTraits({
  classTypes = {
    "harbor", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "necromancer"           ] = defaultTraits({
  classTypes = {
    "cult", },
  faiths = {
    "cult of the damned" },
  ages = npcg.data.age.olderThan( "adult" ),
	alignments = {
		"chaotic evil",
		"neutral evil"
	}
}),
[ "nurse"                 ] = defaultTraits({
  classTypes = {
    "medical", },
  ages = npcg.data.age.olderThan( "teen" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("good"),
}),
-- o
[ "old clothes dealer"    ] = defaultTraits({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "outlaw"                ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "child" ),
	alignments = {
		"lawful evil",
		"chaotic evil",
		"lawful neutral",
		"true neutral",
		"chaotic neutral",
	}
}),
-- p
[ "page"                  ] = defaultTraits({
  classTypes = {
    "paladin", 
    "apprentice",
		"military", },
  races = {
    "blood elf",
    "dwarf",
    "draenei",
    "goblin",
    "half elf",
    "high elf",
    "human",
    "night elf",
    "tauren" },
  ages = npcg.data.age.agesBetween( "child", "child" )
}),
[ "painter"               ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "paladin"               ] = defaultTraits({
  classTypes = {
    "paladin",
		"military", },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "tauren" },
  faiths = npcg.data.faith.faithsByTag( "light" ),
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful good",
		"neutral good",
	}
}),
[ "peasant"               ] = defaultTraits({
  classTypes = {
    "citizen", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "pharmacist"            ] = defaultTraits({
  classTypes = {
    "trader", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "philosopher"           ] = defaultTraits({
  classTypes = {
    "scholar", },
  ages = npcg.data.age.olderThan( "adult" )
}),
[ "physician"             ] = defaultTraits({
  classTypes = {
    "medical", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("good"),
}),
[ "pickpocket"            ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.agesBetween( "child", "teen" ),
	alignments = {
		"chaotic good",
		"neutral good",
		"chaotic neutral",
		"true neutral",
		"chaotic evil",
		"neutral evil",
	}
}),
[ "pie seller"            ] = defaultTraits({
  classTypes = {
    "trader", },   
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "piper"                 ] = defaultTraits({
  classTypes = {
    "crafter", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "tauren",
    "troll",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "piper apprentice"      ] = defaultTraits({
  classTypes = {
    "crafter",
    "apprentice" },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "tauren",
    "troll",
    "worgen" },
  ages = npcg.data.age.agesBetween( "child", "teen" )
}),
[ "pirate"                ] = defaultTraits({
  classTypes = {
    "harbor",
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "teen" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("evil"),
}),
[ "potter"                ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "potter apprentice"     ] = defaultTraits({
  classTypes = {
    "crafter", 
    "apprentice", },
  ages = npcg.data.age.agesBetween( "child", "teen" )
}),
[ "preacher"              ] = defaultTraits({
  classTypes = {
    "priest", 
    "light", },
  ages = npcg.data.age.olderThan( "teen" ),
	alignments = npcg.data.alignment.alignmentsByEthics("lawful"),
}),
[ "priest"                ] = defaultTraits({
  classTypes = {
    "priest",
    "light", },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" ),
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
	}
}),
-- q
-- r
[ "ranger"                ] = defaultTraits({
  classTypes = {
    "hunter",
		"military", },
  races = {
    "blood elf",
		"forsaken",
    "half elf",
		"high elf",
		"void elf", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "rascal"                ] = defaultTraits({
  classTypes = {
    "children", },
  ages = npcg.data.age.agesBetween( "child", "child" )
}),
[ "rebel"                 ] = defaultTraits({
  classTypes = {
    "mercenary", },   
  ages = npcg.data.age.olderThan( "child" ),
	alignments = npcg.data.alignment.alignmentsByEthics("chaotic"),
}),
[ "researcher"						] = defaultTraits({
	races = npcg.data.race.allRacesExcept("death knight"),
	ages = npcg.data.age.olderThan("young adult"),
}),
[	"rift warden"						] = defaultTraits({
	races = {
		"void elf" },
	classTypes = {
		"warlock",
		"scholar", },
}),
[	"riftblade"							] = defaultTraits({
	races = {
		"void elf", },
	classTypes = {
		"military",
		"warrior", },
}),
-- s
[ "sailor"                ] = defaultTraits({
  classTypes = {
    "harbor", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "scholar"               ] = defaultTraits({
  classTypes = {
    "scholar", },
  ages = npcg.data.age.olderThan( "teen" ),
	alignment = npcg.data.alignment.alignmentsByEthics("lawful"),
	races = {
		"blood elf",
		"forsaken",
		"high elf",
		"human",
		"nightborne",
		"pandaren",
		"void elf",
		"worgen",	},
}),
[ "scout"                 ] = defaultTraits({
  classTypes = {
    "hunter",
    "mercenary",
    "military", },   
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" ),
	races = { "any" }
}),
[ "scribe"                ] = defaultTraits({
  classTypes = {
    "scholar", },
  races = npcg.data.race.racesByType( "civilized" ),  
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "scullion"              ] = defaultTraits({
  classTypes = {
    "crafter", },
  races = {
    "blood elf",
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "forsaken",
    "worgen" },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "servant"               ] = defaultTraits({
  classTypes = {
    "citizen", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "sewer"                 ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "shadow hunter"         ] = defaultTraits({
  classTypes = {
    "dark",
		"military", },
  races = {
    "troll" },
  ages = npcg.data.age.olderThan( "adult" ),
	tribes = {
		"darkspear",
		"amani",
		"farraki",
	}
}),
[ "shaman"                ] = defaultTraits({
  classTypes = {
    "shaman", },
  ages = npcg.data.age.agesBetween( "young adult", "venerable" ),
	alignments = {
		"lawful good",
		"neutral good",
		"lawful neutral",
	}
}),
[	"sharpshooter"					] = defaultTraits({
	races = {
		"blood elf",
		"high elf",
		"human",
		"forsaken",
		"worgen",
		"dwarf",
		"troll",
		"goblin",
		"gnome",
		"orc",
		"night elf",
		"void elf",
	},
	tribes = {
		"any",
	},
	classTypes = {
		"hunter",
		"military",
	}
}),
[ "shepherd"              ] = defaultTraits({
  classTypes = {
    "farmer" },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "shopkeeper"            ] = defaultTraits({
  classTypes = {
    "trader", },
  ages = npcg.data.age.agesBetween( "young adult" , "mature adult" )
}),
[ "skinner"               ] = defaultTraits({
  classTypes = {
    "gatherer",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "skinner apprentice"    ] = defaultTraits({
  classTypes = {
    "apprentice",
    "hunter",
    "trader", },
  ages = npcg.data.age.agesBetween( "child", "teen" )
}),
[ "soldier"               ] = defaultTraits({
  classTypes = {
    "military",
    "hunter",
    "paladin", },
  ages = npcg.data.age.agesBetween( "teen", "mature adult" ),
	races = { "any" },
	
}),
[ "sorcerer"              ] = defaultTraits({
  classTypes = {
    "mage",
    "warlock", },   
  ages = npcg.data.age.olderThan("young adult"),
	alignments = {
		"chaotic good",
		"chaotic neutral",
		"chaotic evil",
		"neutral good",
		"true neutral",
		"neutral evil",
	}
}),
[ "spellbreaker"          ] = defaultTraits({
  classTypes = {
    "town support",
    "military", },
  races = {
    "blood elf" },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = {
		"lawful good",
		"lawful neutral",
	}
}),
[ "spy"                   ] = defaultTraits({
  classTypes = {
    "rogue",
    "covert", },
  races = npcg.data.race.allRacesExcept(
    "draenei",
    "tauren"),
  ages = npcg.data.age.olderThan("child"),
	alignments = npcg.data.alignment.alignmentsByEthics("chaotic")
}),
[ "squire"                ] = defaultTraits({
  classTypes = {
    "apprentice", 
    "paladin",
		"military", },
  races = {
    "blood elf",
    "draenei",
    "goblin",
    "half elf",
    "high elf",
    "human",
    "night elf",
    "tauren" },
  ages = npcg.data.age.agesBetween( "child", "young adult" ),
	alignments = {
		"lawful good",
		"lawful neutral",
	}
}),
[ "stablekeeper"          ] = defaultTraits({
  classTypes = {
    "animal",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "stablehand"            ] = defaultTraits({
  classTypes = {
    "animal",
    "hunter" },
  ages = npcg.data.age.agesBetween( "child", "young adult" )
}),
[ "student"               ] = defaultTraits({
  classTypes = {
    "children",
    "scholar",
    "priest",
    "mage" },
  ages = npcg.data.age.agesBetween( "child", "teen" )
}),
[ "surgeon"               ] = defaultTraits({
  classTypes = {
    "medical" },
  races = npcg.data.race.racesByType( "civilized" ), 
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("good"),
}),
[ "swindler"              ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "child" ),
	alignments = {
		"neutral evil",
	}
}),
-- t
[ "tailor"                ] = defaultTraits({
  classTypes = {
    "crafter",
    "trader", },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "teacher"               ] = defaultTraits({
  classTypes = {
    "children",
    "scholar", },
  ages = npcg.data.age.olderThan( "young adult" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("good")
}),
[ "thief"                 ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.agesBetween( "child", "mature adult" ),
	alignments = {
		"chaotic evil",
		"neutral evil",
		"chaotic neutral",
		"true neutral"
	}
}),
[ "thug"                  ] = defaultTraits({
  classTypes = {
    "rogue",
    "illegal", },
  ages = npcg.data.age.olderThan( "teen" ),
	alignments = {
		"chaotic evil",
		"neutral evil",
		"lawful evil",
		"chaotic neutral",
		"true neutral",
		"lawful neutral",
	}
}),
[ "tinker"                ] = defaultTraits({
  classTypes = {
    "crafter",
    "inventor", },
  races = {
    "gnome",
    "goblin" },
  ages = npcg.data.age.olderThan( "young adult" )
}),
[ "trainer"								] = defaultTraits({
	races = npcg.data.race.allRacesExcept("death knight"),
	ages = npcg.data.age.olderThan("adult"),
}),
[ "trapper"               ] = defaultTraits({
  classTypes = {
    "gatherer",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" )
}),
-- u 
-- v
[ "vendor"                ] = defaultTraits({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "veteran"								] = defaultTraits({
	classTypes = {
		"military",
	},
	ages = npcg.data.age.olderThan("adult"),
	races = npcg.data.race.allRacesExcept("death knight", "half elf"),
}),
[	"vindicator"						] = defaultTraits({
	classTypes = {
		"paladin",
		"military",
		"light",
	},
	races = {
		"draenei",
		"troll",
	},
	tribes = {
		"lightforged",
		"",
		"zandalari",
	},
}),
-- w
[ "waiter"                ] = defaultTraits({
  classTypes = {
    "bar", },
  ages = npcg.data.age.agesBetween( "child", "mature adult" )
}),
[ "warlock"               ] = defaultTraits({
  classTypes = {
    "warlock",
    "dark" },   
  ages = npcg.data.age.agesBetween( "young adult",  "venerable" ),
	alignments = npcg.data.alignment.alignmentsWithMoral("evil"),
}),
[ "watchman"              ] = defaultTraits({
  classTypes = {
    "town support",
    "hunter", },
  ages = npcg.data.age.olderThan( "teen" ),
	races = { "any" },
}),
[ "water vendor"          ] = defaultTraits({
  classTypes = {
    "trader", },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "wine vendor"           ] = defaultTraits({
  classTypes = {
    "trader", },
  races = {
    "blood elf",
    "draenei",
    "dwarf",
    "gnome",
    "goblin",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "night elf",
    "pandaren",
    "forsaken" },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "witch doctor"          ] = defaultTraits({
  classTypes = {
    "nature" },
  races = {
    "troll" },
  ages = npcg.data.age.olderThan( "adult" ),
	alignments = {
		"lawful evil",
		"lawful neutral",
		"true neutral",
		"neutral evil",
	}
}),
[ "woodcarver"            ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.olderThan( "child" )
}),
[ "woodworker"            ] = defaultTraits({
  classTypes = {
    "crafter", },
  ages = npcg.data.age.agesBetween( "young adult", "mature adult" )
}),
[ "woolseller"            ] = defaultTraits({
  classTypes = {
    "trader", },
  races = {
    "dwarf",
    "gnome",
    "half elf",
    "half elf",
    "high elf",
    "human",
    "forsaken",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" )
}),
[ "wool-stapler"          ] = defaultTraits({
  classTypes = {
    "crafter", },
  races = {
    "dwarf",
    "gnome",
    "half elf",
    "half orc",
    "high elf",
    "human",
    "forsaken",
    "worgen" },
  ages = npcg.data.age.olderThan( "teen" )
}),
-- x
-- y
-- z
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Data Modifiers | -----------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions allow for customization of defaultTable

-- default :: { Class, ClassConstraints } -> { Class, ClassConstraints }
-- Returns the default class distribution template altered by entries within the given table.
local function default(replacements)
	local replacements = replacements or { }
	
	return functua.table.mapKV(
		function(class, classConstraints)
			return replacements[class] or functua.table.clone(classConstraints) end,
		defaultTable)
end

-- defaultFromWeightTable :: { Class, Int } -> { Class, ClassConstraints }
-- Returns the default class distribution template altered by a weight table. The default weight is 1.
-- Note: If replacements contains anything not present in the default setting, it will not be returned.
local function defaultFromWeightTable(replacements)
	local output = default({ })
	
	for class, value in pairs(replacements) do
		if output[class] ~= nil then
			output[class].weight = value
		end
	end
	
	return output
end

-- defaultFromClassTypes :: { classType, weight } --> { Class, ClassConstraints }
-- Returns the default class distribution template, altered such that given classTypes occur more frequently than those not given.
local function defaultFromClassTypes(replacements)

	local output = default({ })
	
	local findMax = function(op, xs) -- basically a foldr on k, v, aV with a default.
		local outputV = 1 -- default for weights in classes.
		
		for k,v in pairs(xs) do
			if op(k, v, outputV) then
				outputV = v
			end
		end
		return outputV
	end
	
	return functua.table.map(
		function(class)
			class.weight = findMax(
				function(k, v, aV)
					return v >= aV and functua.table.contains(k, class.classTypes) end,
				replacements)			
			return class end,
		output)
end

-- applyReplacementWeights :: { Class, ClassConstraints } -> { Class, Int }
-- Modifies the weights in a given template. (NOT PURE!)
local function applyReplacementWeights(template, replacements)
	local output = template
	for k,v in pairs(replacements) do
		if output[k] ~= nil then
			output[k].weight = v
		end
	end
	return output
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Active Data Templates | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- The default class distribution template, this is called when there are no specifics for zones and/or subzones.
local classes = default({ })

---------------------------------------------------------------------------------------------------------------------------------------
-- | Queries | ------------------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- UNSURE IF LEGACY
-- classesByType :: ClassType -> { Class }
-- Returns all classes of a certain class type.
function data.classesByType(classType)
	--print("classesByType called")
  return functua.table.filterKeysByValue(
    function(x) return
      functua.table.contains(classType, x.classTypes)
    end,
    util.eitherOr(classType == "any", { }, classes)
  )
end

-- UNSURE IF LEGACY!
-- classesByType2 :: ClassType -> { Class, ClassConstraints }
-- Returns all classes and their constraints of a certain class type.
function data.classesByType2(classType)
	--print("classesByType2 called")
  -- this should be horrible?
  local classesWithClassType = data.classesByType(classType)
  local output = { }
  for _,v in ipairs(classesWithClassType) do -- Or is it not..? I can't wrap my head around it >.<
    output[v] = classes[v]
  end
  return output
end

--[[
-- classesByType :: ClassType -> ... -> { Class }
-- Returns all classes satisfying certain class types.
function data.classesByAnyOfTypes(...)
  return functua.mergeTablesWithoutDuplicates(unpack(funtua.map(classesByType, {...})))
end

-- classesByType :: ClassType -> ... -> { Class, ClassConstraints }
-- Returns all classes and their constraints satisfying certain class types.
function data.classesByAnyOfTypes2(...)
  return functua.mergeTablesWithoutDuplicates2(unpack(functua.map(classesByType2, {...})))
end
]]

-- allClassesExcept :: Class -> ... -> { Class }
function data.allClassesExcept(...)
  local exceptions = {...}
  return functua.table.without(functua.table.keys(classes),exceptions)
end

-- randomClass :: () -> Class
-- Generates a true random class without constraints.
-- This is a fallback that should not be used all too lightly.
function data.randomClass()
  return util.randomFrom(functua.table.keys(classes))
end

-- hasAnyOfClassTypes :: ClassTypes a => { Class, a } -> { ClassTypes } -> bool
-- Returns whether the given class has any of the classtypes given.
local function hasAnyOfClassTypes(class, classTypes)
	return functua.table.containsAny(class.classTypes, classTypes)
end

-- getClassesByTypes :: ClassTypes a -> { ClassTypes } -> { Class, a }
function getClassesByTypes(classTypes)
	return functua.table.filterV(function(c) return hasAnyOfClassTypes(c, classTypes) end, classes)
end

---------------------------------------------------------------------------------------------------------------------------------------
-- | Zone-based Generation | ----------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local zonesToClasses = {
  ["dalaran"              ] = nil,
  -- Order Halls
  ["dreadscar rift"       ] = getClassesByTypes({	"warlock" }),
  ["hall of the guardian" ] = getClassesByTypes({ 	 "mage" }),
  ["netherlight temple"   ] = getClassesByTypes({  "priest" }),
  ["skyhold"              ] = getClassesByTypes({ "warrior" }),
  ["the dreamgrove"       ] = getClassesByTypes({   "druid" }),
  ["trueshot lodge"       ] = getClassesByTypes({  "hunter" }),
  
  -- Starter Zones
  ["eversong woods"       ] = nil,
	
	
	-- Orgrimmar
	["miwana's longhouse"		] = nil,

	["engineering works"		] = nil,
	
		-- Warfronts
	[ "arathi highlands"		] = getClassesByTypes({ "military", "covert", "diplomacy", "medical" }),
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Subzone-based Generation | -------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

local subzonesToClasses = {
	-- Horde Cities
  [ "silvermoon city" 		] = { -- Silvermoon City will be the testing playground.
    ["murder row"         						] = getClassesByTypes({ "warlock", "rogue", "illegal" }),
		["the sanctum"										] = getClassesByTypes({ "warlock", "rogue", "illegal" }), -- Warlock Trainer
		
		-- The Bazaar
		["the bazaar"											] = getClassesByTypes({ "crafter", "trader", "town support", "citizen", }),
		["silvermoon finery"  						] = nil, -- Cloth shop
		["keelen's trustworthy tailoring"	] = nil, -- Tailor
		["the sanctum"										] = nil, -- Leper Gnome place under Tailor
		["blades & axes"									] = nil,
		["general goods"									] = nil,
		["shields of silver"							] = nil,
		["the bank of silvermoon"					] = nil,
		["wayfarer's rest"								] = getClassesByTypes({ "bar", }),
		
		-- Walk of Elders
		["walk of elders"									] = nil,
		["velaani's arcane goods"					] = nil,
		["the shepherd's gate"						] = nil,
		["silvermoon registry"						] = nil,
		
		-- The Royal Exchange
		["the royal exchange"							] = nil,
		["the hall of respite"						] = nil,
		["silvermoon jewelry"							] = nil,
		["royal exchange auction house"		] = nil,
		["royal exchange bank"						] = nil,
		["the relinquary"									] = nil,
		["silvermoon city inn"						] = nil,
		
		-- Farstriders' Square
		["farstriders' square"						] = defaultFromClassTypes({
			["paladin"  	] = 200,
			["hunter"   	] = 200,
			["military"		] = 200, }),
		["the hall of blood"							] = applyReplacementWeights(defaultFromClassTypes({
			-- Paladin ClassType
			["paladin"		] = 1000, }), {
			-- Paladin Class.
			["bloodknight"] = 5000, }),
		["the rangers' lodge"							] = getClassesByTypes({ "hunter", }),
		
		-- Court of the Sun
		["court of the sun"								] = nil,
		["sunfury spire"									] = nil,
		["inner sanctum"									] = nil,
		
  },
	[ "eversong woods"  		] = {
		-- Ruins of Silvermoon.
			-- Big house in the back of the Ruins.
			["commons hall"										] = defaultFromClassTypes({
				["illegal"		] = 200,
				["rogue"      ] = 200,
				["warlock"		] = 20, }),
			["dawning lane"										] = defaultFromClassTypes({
				["illegal"		] = 200,
				["rogue"      ] = 200,
				["warlock"		] = 20, }),
			-- Right Road in Ruins.
			["feth's way"											] = defaultFromClassTypes({
				["illegal"		] = 200,
				["rogue"      ] = 200,
				["warlock"		] = 20, }),
			["falconwing square"							] = nil,
			["ruins of silvermoon"						] = defaultFromClassTypes({
				["illegal"		] = 200,
				["rogue"      ] = 200,
				["warlock"		] = 20, }),
			-- Left Road in Ruins.
			["skulking row"										] = defaultFromClassTypes({
				["illegal"		] = 200,
				["rogue"      ] = 200,
				["warlock"		] = 20, }),
	},
	
	-- Orgrimmar
	[ "orgrimmar"						] = {
		["gates of orgrimmar"							] = nil,
		["grommash hold"									] = nil,
		["the drag"												] = nil,
		["valley of strength"							] = nil,
		["valley of wisdom"								] = nil,
		["the western earthstrine"				] = nil,
		["orgrimmar embassy"							] = nil,
		["valley of spirits"							] = nil,
		["cleft of shadow"								] = nil,
		["valley of honor"								] = nil,
		["hall of the brave"							] = nil,
	},
	[	"valley of strength"	] = {
		["the broken tusk"								] = nil,
	},
	[	"valley of honor"			] = {
		["the wyvern's tail"							] = nil,
	},
	-- End of Orgrimmar
	
	-- Alliance Cities
	[	"stormwind city"			] = {
		["valley of heroes"								] = nil,
		
		["trade district"									] = nil,
		["gryphon roost"									] = nil,
		
		["the canals"											] = nil,
		["cordell's enchanting"						] = nil, -- Shop in the Canals, near Mage Quarter
		["the scribe of stormwind"				] = nil,
		
		["the slaughtered lamb"						] = nil,
		["wizard's sanctum"								] = nil,
		
		["lion's rest"										] = nil,
		
		["cathedral square"								] = nil,
		["cathedral of light"							] = nil,
		
		["dwarven district"								] = nil,
		["cut-throat alley"								] = nil,
		
		["stormwind embassy"							] = nil,
		["olivia's pond"									] = nil,
		["the wollerton stead"						] = nil,
		["the eastern earthstrine"				] = nil,
		["stormwind lake"									] = nil,
		["stormwind city cemetery"				] = nil,
		
		["stormwind keep"									] = nil,
		["petitioner's chamber"						] = nil,
		["royal library"									] = nil,
		
		["old town"												] = nil,
		["champion's hall"								] = nil,
		["si:7"														] = nil,
		["command center"									] = nil,
		
		["stormwind harbor"								] = nil,
	},
	[	"the canals"					] = {
		["dwarven district"								] = nil,
	},
	[	"old town"						] = {
		["pig and whistle tavern"					] = nil,
	},
	
	-- Dalaran (both versions)
  [ "dalaran"         		] = {
    -- Rogue Order Hall
    ["chamber of shadows" 						] = data.classesByType2("rogue"),
    ["cloaked vestibule"  						] = data.classesByType2("rogue"),
    ["hidden corridor"    						] = data.classesByType2("rogue"),
    ["the hall of shadows"						] = data.classesByType2("rogue"),
    ["uncrowned vault"    						] = data.classesByType2("rogue"),
  },
}

---------------------------------------------------------------------------------------------------------------------------------------
-- | Functions for Generation and Retrieval | -----------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- Main generator should be CharacterTemplate -> { Trait }
-- This will later change to be CharacterTemplate -> { Trait, weight } or something similar.

-- Part 1: Get potential candidates
-- Part 2: Weed out unfeasable candidates, like those not within context or faction
-- Part 3: Select a random item

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 1 Functions: Generation | -------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of the type:
-- characterTemplate -> zone -> subZone -> { class, classconstraints }

-- If we change the way zones and subzones work in the dataset, and make them a tree, we can abstract these functions.

local function checkAllLocations(characterTemplate, zone, subZone)
  return 
    -- Test whether the SubZone has values.
    ((subzonesToClasses[zone] or { })[subZone])
    -- Test whether the Zone has values.
    or zonesToClasses[zone]
    -- Fallback to defaultTraits.
    or classes
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function checkZonesOnly(characterTemplate, zone)
  return
    -- Test whether the Zone has values.
    zonesToClasses[zone]
    -- Fallback to defaultTraits.
    or classes
    -- Fallback in case some extreme breaking happened.
    or { }
end

local function dontCheckLocations(characterTemplate)
  return
    -- Fallback to defaultTraits.
    classes
    -- Fallback in case some extreme breaking happened.
    or { }
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 2 Functions: Filtering | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- characterTemplate -> { class, classconstraints } -> zone -> subzone -> bool

-- We can rework this one to take in a dependency list.

local function meetsTraitConstraints(characterTemplate, option)
	return  functua.table.containsAny(option.races,   		{ "any", characterTemplate.race   		})
    and   functua.table.containsAny(option.ages,    		{ "any", characterTemplate.age    		})
    --and   functua.table.containsAny(option.faiths,  		{ "any", characterTemplate.faith  		})
		--and		functua.containsAny(option.alignments, 	{ "any", characterTemplate.alignment 	})
		and		functua.table.containsAny(option.genders,			{ "any", characterTemplate.gender			})
		and		functua.table.containsAny(option.tribes,			{ "any", characterTemplate.tribe			})
end

local function isWowClass(characterTemplate, option)
  return functua.table.any(
    functua.composeEqualsFunction(option),
    { "druid", "hunter", "mage", "monk", "paladin", "priest", "rogue", "shaman", "warlock", "warrior" })
end

---------------------------------------------------------------------------------------------------------------------------------------
---- | Part 3 Functions: Selection | --------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- These functions should be of type:
-- Arbitrary a => { class, a } -> class
-- This might change in the future.

local function randomKeyFrom(options)
  return util.randomFrom(functua.table.keys(options))
end

local function selectUsingWeightedRandom(selectionList)
  return util.weightedRandomFrom(functua.table.keys(selectionList), function(item) return selectionList[item].weight end)
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Combination Function | ------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------

-- createClassGenerator :: Trait a, TraitConstraints b =>
--	(characterTemplate -> zone -> subZone -> { a, b }) -> 
--	(characterTemplate -> { a, b } -> zone -> subzone -> bool) -> 
--	({ d, c } -> d)
--	-> (characterTemplate, zone, subZone) -> a

function createClassGenerator(generatorF, filterF, selectorF)
  return function(characterTemplate, zone, subZone)
    return selectorF(
      functua.table.filterV( -- Returns KVP, filters on V.
        function(potentialTrait) return filterF(characterTemplate, potentialTrait, zone, subZone) end,  -- Most don't use zone or subzone here.
        generatorF(characterTemplate, zone, subZone)
      )
    )
  end
end

local traitAndSubZoneBasedClassGenerator          = createClassGenerator(checkAllLocations,   meetsTraitConstraints,  selectUsingWeightedRandom)
local traitAndZoneBasedClassGenerator             = createClassGenerator(checkZonesOnly,      meetsTraitConstraints,  selectUsingWeightedRandom)
local traitAndNoLocationBasedClassGenerator       = createClassGenerator(dontCheckLocations,  meetsTraitConstraints,  selectUsingWeightedRandom)
local wowOnlyAndSubZoneBasedClassGenerator        = createClassGenerator(checkAllLocations,   isWowClass,             selectUsingWeightedRandom)
local wowOnlyAndZoneBasedClassGenerator           = createClassGenerator(checkZonesOnly,      isWowClass,             selectUsingWeightedRandom)
local wowOnlyAndNoLocationBasedClassGenerator     = createClassGenerator(dontCheckLocations,  isWowClass,             selectUsingWeightedRandom)
local noContextAndSubZoneBasedClassGenerator      = createClassGenerator(checkAllLocations,   functua.const(true),    selectUsingWeightedRandom)
local noContextAndZoneBasedClassGenerator         = createClassGenerator(checkZonesOnly,      functua.const(true),    selectUsingWeightedRandom)
local noContextAndNoLocationBasedClassGenerator   = createClassGenerator(dontCheckLocations,  functua.const(true),    selectUsingWeightedRandom)

function generator.initialize()
  generateRandomClass = function() return "Error: generateRandomClass is not set" end
  
  if not NPCG_Options.locationSettings.zones then
    generateRandomClass = function(characterTemplate)
      return traitAndNoLocationBasedClassGenerator(characterTemplate)
      or noContextAndNoLocationBasedClassGenerator(characterTemplate) end
  elseif NPCG_Options.locationSettings.subZones then
    generateRandomClass = function(characterTemplate, zone, subZone)
      return traitAndSubZoneBasedClassGenerator(characterTemplate, zone, subZone)
      or noContextAndSubZoneBasedClassGenerator(characterTemplate, zone, subZone)
    end
  else
    generateRandomClass = function(characterTemplate, zone)
      return traitAndZoneBasedClassGenerator(characterTemplate, zone)
      or noContextAndZoneBasedClassGenerator(characterTemplate, zone)
    end
  end
end

---------------------------------------------------------------------------------------------------------------------------------------
--| Testing Functions | ---------------------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------------------------------------------


function tests.testZoneToClassRatio()
  local test = { }
  local zone   =  util.eitherOr(
                    NPCG_Options.locationSettings.mock  and 
                    NPCG_Options.locationSettings.zones, 
                    NPCG_Options.locationSettings.mockZone,
                    GetZoneText())
  local subzone = util.eitherOr(
                    NPCG_Options.locationSettings.mockSub and
                    NPCG_Options.locationSettings.zones   and
                    NPCG_Options.locationSettings.subZones,
                    NPCG_Options.locationSettings.mockSubZone or "<No SubZone>",
                    GetSubZoneText() or "<No SubZone>")
  print("Testing Zone to Class Ratio for: " .. zone .. ", " .. subzone)
  functua.forEach(
    function(x) 
      rollAll() -- Roll a full npc
			result = GetCurrentCharacterTemplate()
      test[result.class] = (test[result.class] or 0) + 1 end,
    functua.replicate(500, 1))
  util.logTableMapped(function(x) return x .. ", (" .. ((x / 500) * 100) .. "%)" end, test)
end


-- LEGACY
function tests.testClassesByType2()
  print "Testing classesByType2(\"rogue\")"
  local results = data.classesByType2("rogue")
  print "Keys"
  util.logTable(functua.keys(results))
  print "Values"
  functua.map(util.logTableOfPropertyLists, results)
  print "Ended testing"
end

-- End of File.